const Jimp = require('jimp');
const { GifUtil } = require('gifwrap');
const trueTo256 = require('./trueTo256');

// 需要添加的水印图片路径
const ORIGINAL_IMAGE = './img/test.gif';

// 水印logo路径
const LOGO = './img/logo.png';

// 水印距离右下角百分比
const LOGO_MARGIN_PERCENTAGE = 5 / 100;

async function main () {
  const logo = await Jimp.read(LOGO);

  return GifUtil.read(ORIGINAL_IMAGE).then(inputGif => {
    // 将 logo 等比缩小 10 倍
    // logo.resize(inputGif.width / 10, Jimp.AUTO);

    const xMargin = inputGif.width * LOGO_MARGIN_PERCENTAGE;
    const yMargin = inputGif.height * LOGO_MARGIN_PERCENTAGE;

    const X = inputGif.width - logo.bitmap.width - xMargin;
    const Y = inputGif.height - logo.bitmap.height - yMargin;

    // 给每一帧都打上水印
    inputGif.frames.forEach((frame, i) => {
      // 只为第一帧添加水印，可能会出现水印被覆盖问题
      /* if (i !== 0) {
        return;
      } */
      const jimpCopied = GifUtil.copyAsJimp(Jimp, frame);

      // 计算获得的坐标再减去每一帧偏移位置，为实际添加水印坐标
      jimpCopied.composite(logo, X - frame.xOffset, Y - frame.yOffset, [{
        mode: Jimp.BLEND_SOURCE_OVER,
        opacitySource: 0.1,
        opacityDest: 1
      }]);

      // 压缩图片 0-100
      jimpCopied.quality(80);

      frame.bitmap = jimpCopied.bitmap;

      // 输出每一帧图片
      // jimpCopied.write(`${i}.png`);

      // 真彩色转 256 色
      frame.bitmap = trueTo256(frame.bitmap);
    });

    // 不使用 trueTo256 也可以使用自带的 quantizeWu 进行颜色转换，不过自带的算法运行需要更多的时间，没有 trueTo256 快
    // GifUtil.quantizeWu(inputGif.frames);

    return inputGif;
  });
}

main().then(inputGif => {
  // Pass inputGif to write() to preserve the original GIF's specs.
  const FILENAME = 'new_name.gif';
  return GifUtil.write(FILENAME, inputGif.frames, inputGif).then(outputGif => {
    console.log('水印成功：', FILENAME);
  }).catch((err) => {
    if (err) {
      return console.error('水印失败：', err);
    }
  });
});
